	TITLE	'character I/O'
; character I/O drivers for CP/M 3 BIOS (C) Duncan Elvin

	NAME	'CHARIO'
	MACLIB	MODEBAUD

	PUBLIC	?CINIT,?CI,?CO,?CIST,?COST,?CST
	PUBLIC	@CTBL

NDEV	EQU	4			; Number of devices

; initialise specified device
?CINIT:	MOV	A,C			; device number
	CPI	NDEV			; in range?
	RNC
	ADD	A			; * 8
	ADD	A
	ADD	A
	MOV	L,A
	MVI	H,0
	LXI	D,CRTBD			; index into character device table
	DAD	D
	MOV	A,M			; get baud rate
	ANA	A			; if baud rate = 0, don't initialise
	RZ
	MOV	B,C			; device number
	LXI	H,CINI			; address of pointer to jump table
	JMP     VECTOR

; character input
?CI:	
	LXI	H,CIN			; address of pointer to jump table
	JMP    	VECTOR

; character input status
?CIST:	LXI	H,CIS			; address of pointer to jump table
	JMP	VECTOR

; special input status for CONIN keyboard always returns true
?CST:	LXI	H,CS
	JMP	VECTOR

; character output
?CO:	MOV	A,C			; character to output
	LXI	H,COUT			; address of pointer to jump table
	JMP	VECTOR

; character output status
?COST:	LXI	H,COS			; address of pointer to jump table
	JMP	VECTOR

; vector via jump tables
; given:    HL = address of address of start of jump table
;           B = device number
;           A = argument to pass
; returned: A = value returned from routine

VECTOR: PUSH    A                                      
        MOV     A,B
	ADD     B
	ADD     B                 
        MOV     E,A
        MVI     D,0
        DAD     D             
        POP     A 
        PCHL

;
@CTBL:	DB	'SIO   '		; device 0 is RS232 Port
	DB	MB$IN$OUT+MB$SOFTBAUD+MB$SERIAL
CRTBD:	DB	BAUD$9600                       
	DB	'LPT   '		; device 1 is Centronics Printer Port
	DB	MB$OUTPUT
	DB	BAUD$NONE
	DB	'KBD   '		; device 2 is Keyboard
	DB	MB$INPUT
	DB	BAUD$NONE
	DB	'VDP   '                ; device 3 is Video Display Processor
	DB	MB$OUTPUT
	DB	BAUD$NONE
        DB      0			; end of device table

CINI:   JMP     SET$BAUD		; device init table
        RET ! NOP ! NOP
        RET ! NOP ! NOP
	RET ! NOP ! NOP

CIS:    JMP     SIO$IPS                ; input status table    
        RET ! NOP ! NOP
        JMP     KBD$ST
        RET ! NOP ! NOP

CS:     JMP     SIO$IPS                ; special input status table    
        RET ! NOP ! NOP
        MVI A,0FFH ! RET
        RET ! NOP ! NOP

CIN:    IN 10H ! RET			; input table
        RET ! NOP ! NOP   
        JMP     KBD$IN
        RET ! NOP ! NOP   
    
COS:    JMP     SIO$OPS                ; output status table
        JMP     LST$ST             
        RET ! NOP ! NOP
 	MVI A,0FFH ! RET

COUT:   OUT 10H	! RET			; output table
	RST 1
	DB 9FH
	RET
        RET ! NOP ! NOP
	RST 1
 	DB 9EH
	RET
   
SCC$OP: MOV B,A		; TO ALLOW ME TO TEST CODE D.ELVIN
SCC0:   IN 81H
        ANI 4
        JZ SCC0
        IN 83H
        CPI 19
        JZ SCC0
        MOV A,B
        OUT 83H
        RET

; CTC BAUD RATE GENERATOR MODE BYTES
T256	EQU	3DH
T16	EQU	1DH
CNTR	EQU	5DH

SET$BAUD:   PUSH A
        MVI A,0B5H
        OUT 11H
        POP A
	ADD	A
	LXI	H,BAUDS
	MOV	C,A
	MVI	B,0
	DAD	B
	MOV	A,M
	INX	H
	MOV	B,M
	MVI	C,28H
	DB 0EDH,79H	; OUT (C),A
	NOP
	DB 0EDH,41H	; OUT (C),B
	INR	C
	DB 0EDH,79H
	NOP
	DB 0EDH,41H
	RET

BAUDS:
; FORMAT: MODE_BYTE,TIME_CONSTANT  ACTUAL BAUD  TARGET BAUD
; ALL ERRORS LESS THAN 1% UNLESS OTHERWISE STATED
	DB 0,0		; NO BAUD
	DB T256,14H	; 49	50 2% ERROR
	DB T16,0D0H	; 75	75
	DB T16,8EH	; 110	110
	DB T16,75H	; 134	134
	DB T16,68H	; 150	150
	DB T16,34H	; 300	300
	DB T16,1AH	; 601	600
	DB CNTR,68H	; 1202	1200
	DB CNTR,45H	; 1812	1800
	DB CNTR,34H	; 2404	2400
	DB CNTR,23H	; 3571	3600
	DB CNTR,1AH	; 4808	4800
	DB CNTR,11H	; 7353	7200 2% ERROR
	DB CNTR,0DH	; 9615	9600
	DB CNTR,07H	; 17857 19200 7% ERROR

KBD$ST:     RST 1
            DB 9BH
            ORA A
            RZ
            XRA A
            DCR A
            RET

KBD$IN:     RST 1
            DB 9CH
            RET

SIO$IPS: IN 11H
         ANI 02H
         RZ
         XRA A
         DCR A
         RET

SIO$OPS: IN 11H
         ANI 01H
         RZ
         XRA A
         DCR A
         RET

LST$ST:  LDA 0FB92H
         ANI 01H
         JNZ L3
         XRA A
L3:      DCR A
         RET
           
         END
T$ST